# -*- coding: utf-8 -*-
"""
Test the functions related to configuring proposal methods
"""
import numpy as np
import os
import pytest
from unittest.mock import MagicMock, patch

from nessai.samplers.nestedsampler import NestedSampler
from nessai.proposal import (
    AnalyticProposal,
    RejectionProposal,
)


@pytest.fixture
def sampler(sampler, tmpdir):
    sampler.output = tmpdir.mkdir("test")
    sampler.plot = False
    sampler.n_pool = None
    sampler.acceptance_threshold = 0.01
    return sampler


@pytest.mark.parametrize(
    "maximum, result", [[None, 200], [False, 0], [100, 100], ["inf", np.inf]]
)
def test_uninformed_maximum(sampler, maximum, result):
    """
    Test to check that the proposal is correctly configured depending on
    the maximum number of uninformed iterations.
    """
    NestedSampler.configure_uninformed_proposal(
        sampler, None, False, maximum, None
    )
    assert sampler.maximum_uninformed == result

    if maximum is False:
        assert sampler.uninformed_sampling is False
    else:
        assert sampler.uninformed_sampling is True


def test_uninformed_threshold(sampler):
    """Test the check uninformed threshold is set correctly"""
    NestedSampler.configure_uninformed_proposal(
        sampler, None, False, None, 0.5
    )
    assert sampler.uninformed_acceptance_threshold == 0.5


def test_uninformed_threshold_default_below(sampler):
    """
    Test to check that the threshold is set to 10 times the acceptance
    if it is below 0.1.
    """
    sampler.acceptance_threshold = 0.05
    NestedSampler.configure_uninformed_proposal(
        sampler, None, False, None, None
    )
    assert sampler.uninformed_acceptance_threshold == 0.5


@pytest.mark.parametrize("threshold", [0.1, 0.2])
def test_uninformed_threshold_default_(sampler, threshold):
    """
    Test to check that the threshold is set to the same value if it is above
    or equal to 0.1
    """
    sampler.acceptance_threshold = threshold
    NestedSampler.configure_uninformed_proposal(
        sampler, None, False, None, None
    )
    assert sampler.uninformed_acceptance_threshold == threshold


def test_uninformed_no_analytic_priors(sampler):
    """
    Test to check that the correct proposal method is used without analytic
    priors.
    """
    NestedSampler.configure_uninformed_proposal(
        sampler, None, False, None, None
    )
    assert isinstance(sampler._uninformed_proposal, RejectionProposal)


def test_uninformed_analytic_priors(sampler):
    """
    Test to check that the correct proposal method is used with analytic
    priors.
    """
    NestedSampler.configure_uninformed_proposal(
        sampler, None, True, None, None
    )
    assert isinstance(sampler._uninformed_proposal, AnalyticProposal)


def test_uninformed_proposal_class(sampler):
    """Test using a custom proposal class"""
    from nessai.proposal.base import Proposal

    class TestProposal(Proposal):
        def draw(self, point):
            pass

    NestedSampler.configure_uninformed_proposal(
        sampler, TestProposal, False, None, None
    )
    assert isinstance(sampler._uninformed_proposal, TestProposal)


def test_configure_flow_proposal(sampler):
    fake_class = MagicMock()
    fake_class.__name__ = "fake_class"
    flow_config = dict(patience=10)
    kwargs = dict(test=True, invalid=True)
    expected_kwargs = dict(test=True, poolsize=sampler.nlive)

    with patch(
        "nessai.samplers.nestedsampler.get_region_sampler_proposal_class",
        return_value=fake_class,
    ) as mock_get, patch(
        "nessai.samplers.nestedsampler.check_proposal_kwargs",
        return_value=expected_kwargs,
    ) as mock_check:

        NestedSampler.configure_flow_proposal(
            sampler, None, flow_config, False, **kwargs
        )

    mock_get.assert_called_once_with(None)
    mock_check.assert_called_once_with(
        fake_class,
        {**kwargs, "poolsize": sampler.nlive},
    )

    fake_class.assert_called_once_with(
        sampler.model,
        flow_config=flow_config,
        output=os.path.join(sampler.output, "proposal", ""),
        plot=False,
        **expected_kwargs,
    )


@pytest.mark.parametrize("val", [(0.1, 10), (0.8, 200)])
def test_proposal_switch(sampler, val):
    """Test the method for switching proposals"""
    sampler.mean_block_acceptance = 0.5
    sampler.mean_acceptance = val[0]
    sampler.uninformed_acceptance_threshold = 0.5
    sampler.iteration = val[1]
    sampler.maximum_uninformed = 100
    sampler._flow_proposal = MagicMock()
    sampler._flow_proposal.n_pool = 2
    sampler._flow_proposal.configure_pool = MagicMock()
    sampler._uninformed_proposal = MagicMock()
    sampler._uninformed_proposal.pool = True
    sampler._uninformed_proposal.close_pool = MagicMock()
    sampler.proposal = sampler._uninformed_proposal

    assert NestedSampler.check_proposal_switch(sampler) is True

    assert sampler.uninformed_sampling is False
    assert sampler.proposal == sampler._flow_proposal


def test_proposal_no_switch(sampler):
    """Ensure proposal is not switched"""
    sampler.mean_acceptance = 0.5
    sampler.uninformed_acceptance_threshold = 0.1
    sampler.iteration = 10
    sampler.maximum_uninformed = 100
    assert NestedSampler.check_proposal_switch(sampler) is False


def test_proposal_already_switched(sampler):
    """Test switching when proposal is already switched"""
    sampler.mean_acceptance = 0.5
    sampler.uninformed_acceptance_threshold = 0.1
    sampler.mean_block_acceptance = 0.5
    sampler.iteration = 10
    sampler.maximum_uninformed = 100
    sampler._flow_proposal = MagicMock()
    sampler._flow_proposal.ns_acceptance = 0.2
    sampler.uninformed_sampling = False
    sampler.proposal = sampler._flow_proposal
    assert NestedSampler.check_proposal_switch(sampler, force=True) is True
    assert sampler.proposal is sampler._flow_proposal
    # If proposal was switched again, acceptance would change
    assert sampler.proposal.ns_acceptance == 0.2



# -*- coding: utf-8 -*-
"""Utilities for proposal classes"""
from copy import deepcopy
from inspect import getmro, signature
import logging


logger = logging.getLogger(__name__)


def check_proposal_kwargs(ProposalClass, kwargs, strict=False):
    """Check the keyword arguments are correct for a proposal class.

    Removes any keyword arguments that correspond to another class. Raises an
    error if the keyword arguments are unknown. For example if the class is
    FlowProposal, it will remove any keyword arguments that correspond to
    AugmentedFlowProposal but not raise an error.

    Parameters
    ----------
    ProposalClass : Type[FlowProposal]
        Class to check the keyword arguments against.
    kwargs : dict
        Dictionary of keyword arguments.
    strict : bool
        Raise an error if any extra keyword arguments are provided even if they
        correspond to other proposals.

    Returns
    -------
    dict
        Dictionary of updated kwargs.
    """
    from ..proposal import AugmentedFlowProposal, FlowProposal
    from ..gw.proposal import AugmentedGWFlowProposal, GWFlowProposal

    proposals = {
        AugmentedFlowProposal,
        AugmentedGWFlowProposal,
        FlowProposal,
        GWFlowProposal,
    }

    class_keys = set()
    for cls in getmro(ProposalClass):
        class_keys.update(signature(cls).parameters.keys())
        if cls in proposals:
            proposals.remove(cls)

    keys = set(kwargs.keys())
    kwargs_out = deepcopy(kwargs)

    extra_keys = keys - class_keys

    if not extra_keys:
        logger.debug("All keyword arguments match the proposal class")
        return kwargs_out
    elif strict:
        raise RuntimeError(
            f"Keyword arguments contain unknown keys: {extra_keys}"
        )

    allowed_extra_keys = set()

    for proposal in proposals:
        allowed_extra_keys.update(set(signature(proposal).parameters.keys()))

    invalid_keys = extra_keys - allowed_extra_keys

    if invalid_keys:
        raise RuntimeError(
            f"Unknown kwargs for {ProposalClass.__name__}: {invalid_keys}."
        )
    else:
        logger.warning(
            f"Removing unused keyword arguments ({extra_keys}) from kwargs for"
            f" {ProposalClass.__name__}. These are valid keyword arguments but"
            " correspond to other proposal classes."
        )
        for key in extra_keys:
            kwargs_out.pop(key)
    return kwargs_out


def get_region_sampler_proposal_class(proposal_class):
    """Get the proposal class for the standard region sampler.

    Parameters
    ----------
    proposal_class : Union[str, Proposal, None]
        The name of the proposal class or the class itself. If not specified,
        defaults to :py:obj:`nessai.proposal.flowpropsoal.FlowProposal`.

    Returns
    -------
    Proposal class
    """
    from .flowproposal import FlowProposal

    if proposal_class is not None:
        if isinstance(proposal_class, str):
            proposal_class = proposal_class.lower()
            if proposal_class == "gwflowproposal":
                from ..gw.proposal import GWFlowProposal as proposal_class
            elif proposal_class == "augmentedgwflowproposal":
                from ..gw.proposal import (
                    AugmentedGWFlowProposal as proposal_class,
                )
            elif proposal_class == "flowproposal":
                proposal_class = FlowProposal
            elif proposal_class == "augmentedflowproposal":
                from ..proposal import AugmentedFlowProposal

                proposal_class = AugmentedFlowProposal
            elif proposal_class == "clusteringflowproposal":
                from ..experimental.proposal.clustering import (
                    ClusteringFlowProposal,
                )

                proposal_class = ClusteringFlowProposal
            elif proposal_class == "clusteringgwflowproposal":
                from ..experimental.gw.proposal import (
                    ClusteringGWFlowProposal,
                )

                proposal_class = ClusteringGWFlowProposal
            else:
                raise ValueError(f"Unknown flow class: {proposal_class}")
        elif not issubclass(proposal_class, FlowProposal):
            raise RuntimeError(
                "Flow class must be string or class that "
                "inherits from FlowProposal"
            )
    else:
        proposal_class = FlowProposal
    return proposal_class



# -*- coding: utf-8 -*-
"""
Test the FlowModel utils.
"""
import pytest
from unittest.mock import patch

from nessai.flowmodel import config as default_config
from nessai.flowmodel.utils import (
    update_config,
    update_flow_config,
    update_training_config,
)


def test_update_config():
    flow_config = dict(n_blocks=2)
    training_config = dict(patience=1)
    expected_out_flow = object()
    expected_out_training = object()
    with patch(
        "nessai.flowmodel.utils.update_flow_config",
        return_value=expected_out_flow,
    ) as update_fcfg, patch(
        "nessai.flowmodel.utils.update_training_config",
        return_value=expected_out_training,
    ) as update_tcfg:

        flow_config_out, training_config_out = update_config(
            flow_config, training_config
        )

    update_fcfg.assert_called_once_with(flow_config)
    update_tcfg.assert_called_once_with(training_config)

    assert flow_config_out is expected_out_flow
    assert training_config_out is expected_out_training


def test_update_config_deprecated():
    """Test with a basic config"""
    model_config = dict(n_blocks=2)
    config = dict(
        patience=1,
        device_tag="cuda",
        inference_device_tag="cuda",
        model_config=model_config,
    )
    flow_config, training_config = update_config(config)
    assert flow_config["n_blocks"] == 2
    assert training_config["patience"] == 1
    assert training_config["device_tag"] == "cuda"
    assert training_config["inference_device_tag"] == "cuda"


def test_update_training_config_invalid_type():
    """Test update config when an invalid argument is specified"""
    with pytest.raises(TypeError, match="Must pass a dictionary"):
        update_training_config(False)


def test_update_training_config_none():
    assert update_training_config(None) == default_config.training.asdict()


def test_update_training_config_invalid_noise_scale():
    """Assert an error is raised if noise_scale is not a float"""
    config = {"noise_scale": "auto", "noise_type": "adaptive"}
    with pytest.raises(TypeError, match=r"`noise_scale` must be a float"):
        update_training_config(config)


def test_update_training_config_missing_noise_type():
    """Assert noise type is set if only `noise_scale` is given"""
    config = {"noise_scale": 1.0}
    out = update_training_config(config)
    assert out["noise_type"] == "constant"


def test_update_training_config_missing_noise_scale():
    """Assert an error is raised if noise_scaling is missing when noise_type is
    set.
    """
    config = {"noise_type": "constant"}
    with pytest.raises(RuntimeError, match="`noise_scale` must be specified"):
        update_training_config(config)


def test_update_flow_config_n_neurons():
    """Assert `get_n_neurons` is called"""
    config = dict(n_inputs=10)
    with patch(
        "nessai.flowmodel.utils.get_n_neurons", return_value=20
    ) as mock:
        config = update_flow_config(config)
    mock.assert_called_once_with(n_neurons=None, n_inputs=10)
    assert config["n_neurons"] == 20


def test_update_model_config_type_error():
    """Assert a type error is raised if the input is the wrong type"""
    with pytest.raises(TypeError, match="Must pass a dictionary"):
        update_flow_config(False)


def test_update_flow_config_none():
    """Test update config for no input"""
    assert update_flow_config(None) == default_config.flow.asdict()



# -*- coding: utf-8 -*-
"""
Default configuration of FlowModel.
"""
from dataclasses import dataclass
from typing import Callable

from ..config import _BaseConfig


@dataclass
class FlowConfig(_BaseConfig):
    """Default configuration for flows."""

    # Use a dataclass since dictionaries are mutable
    n_inputs: int = None
    n_neurons: int = None
    n_blocks: int = 4
    n_layers: int = 2
    ftype: str = "RealNVP"
    flow: Callable = None
    distribution: str = None
    distribution_kwargs: dict = None


@dataclass
class TrainingConfig(_BaseConfig):
    """Default configuration for training."""

    # Use a dataclass since dictionaries are mutable
    device_tag: str = "cpu"
    inference_device_tag: str = None
    lr: float = 0.001
    annealing: bool = False
    clip_grad_norm: float = 5.0
    batch_size: int = 1000
    val_size: float = 0.1
    max_epochs: int = 500
    patience: int = 20
    noise_type: str = None
    noise_scale: float = None
    use_dataloader: bool = False
    optimiser: str = "adamw"
    optimiser_kwargs: dict = None


flow = FlowConfig()
training = TrainingConfig()



import copy
import logging
import numpy as np
from scipy.special import logsumexp
from typing import Any, Tuple

from ...flowmodel.base import FlowModel

logger = logging.getLogger(__name__)


class ClusteringFlowModel(FlowModel):
    """Flow model that includes clustering."""

    max_n_clusters: int = 20
    n_clusters: int = None
    cluster_weights: np.ndarray = None

    def __init__(self, flow_config=None, training_config=None, output=None):
        try:
            import faiss

            logger.debug(f"Running with faiss version {faiss.__version__}")

        except ImportError:
            raise RuntimeError(
                "faiss is not installed! Install faiss-cpu/-gpu in order to "
                "use the clustering flow model."
            )
        super().__init__(
            flow_config=flow_config,
            training_config=training_config,
            output=output,
        )

    def setup_from_input_dict(
        self, flow_config: dict, training_config
    ) -> None:
        flow_config = copy.deepcopy(flow_config)
        if flow_config is None:
            flow_config = {}
        max_n_clusters = flow_config.pop("max_n_clusters", None)
        super().setup_from_input_dict(flow_config, training_config)
        if max_n_clusters is not None:
            self.max_n_clusters = max_n_clusters
        self.flow_config["context_features"] = 1

    def train_clustering(self, samples: np.ndarray, **kwargs) -> np.ndarray:
        """Train the clustering algorithm using a set of samples.

        Tries :code:`n_clusters=2` up to :code:`max_n_clusters`
        and chooses the number that has highest silhouette score.

        Returns
        -------
        Array of labels for the training samples.
        """
        import faiss

        best_score = -np.inf
        best_clusterer = None
        dims = samples.shape[-1]
        for n_clusters in range(2, self.max_n_clusters + 1):
            kmeans = faiss.Kmeans(
                dims, n_clusters, nredo=5, niter=20, **kwargs
            )
            kmeans.train(samples)
            score = np.mean(silhouette_score(samples, kmeans))
            logger.debug(f"k={n_clusters}, score={score}")
            if score > best_score:
                best_score = score
                best_clusterer = kmeans

        if best_clusterer is None:
            raise RuntimeError("Clustering failed")
        self.clusterer = best_clusterer
        labels = self.get_cluster_labels(samples)
        unique_labels = np.unique(labels)
        self.n_clusters = len(unique_labels)
        logger.debug(f"n_clusters={self.n_clusters}")
        self.cluster_weights = np.bincount(
            labels.flatten(), minlength=self.n_clusters
        ) / len(samples)
        logger.debug(f"cluster_weights={self.cluster_weights}")
        return labels

    def get_cluster_labels(
        self, samples: np.ndarray, clusterer=None
    ) -> np.ndarray:
        """Get the cluster labels for a set of samples."""
        if clusterer is None:
            clusterer = self.clusterer
        _, labels = clusterer.index.search(samples, 1)
        return labels.reshape(-1, 1)

    def sample_cluster_labels(self, n: int) -> np.ndarray:
        """Sample n random cluster labels"""
        return np.random.choice(
            self.n_clusters, size=(n, 1), p=self.cluster_weights
        )

    def train(self, samples: np.ndarray, **kwargs) -> dict:
        """Train the normalising flow and clustering"""
        cluster_labels = self.train_clustering(samples)
        return super().train(samples, conditional=cluster_labels, **kwargs)

    def forward_and_log_prob(
        self, x: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:
        cluster_labels = self.get_cluster_labels(x)
        x, _ = super().forward_and_log_prob(x, conditional=cluster_labels)
        log_prob = self.log_prob(x)
        return x, log_prob

    def log_prob(self, x: np.ndarray) -> np.ndarray:
        # Must compute log-prob for every conditional value
        cluster_labels = np.tile(np.arange(self.n_clusters), len(x))[
            :, np.newaxis
        ]
        x = np.repeat(x, self.n_clusters, axis=0)
        log_prob = (
            super()
            .log_prob(x, conditional=cluster_labels)
            .reshape(-1, self.n_clusters)
        )
        return logsumexp(log_prob, b=self.cluster_weights, axis=1)

    def sample(self, n: int = 1, return_labels: bool = False) -> np.ndarray:
        cluster_labels = self.sample_cluster_labels(n)
        samples = super().sample(n, conditional=cluster_labels)
        if return_labels:
            return samples, cluster_labels
        else:
            return samples

    def sample_and_log_prob(
        self, N: int = 1, z: np.ndarray = None, alt_dist: Any = None
    ) -> Tuple[np.ndarray, np.ndarray]:
        if alt_dist is not None:
            raise RuntimeError
        if z is not None:
            N = len(z)
        # This could be optimised to not repeat calculations
        samples = self.sample(N)
        log_prob = self.log_prob(samples)
        return samples, log_prob


def silhouette_score(samples: np.ndarray, clusterer) -> np.ndarray:
    """Compute the silhouette score.

    Based on: https://github.com/facebookresearch/faiss/issues/1875
    """
    distance, _ = clusterer.index.search(samples, 2)
    score = (distance[:, 1] - distance[:, 0]) / np.max(distance, 1)
    return score



from nessai.flowmodel import FlowModel
from nessai.experimental.flows.glasflow import (
    GlasflowWrapper,
    get_glasflow_class,
    known_flows,
)
import numpy as np
import pytest


@pytest.mark.parametrize("name", known_flows.keys())
def test_get_glasflow_class(name):
    FlowClass = get_glasflow_class(f"glasflow-{name}")
    FlowClass(n_inputs=2, n_neurons=4, n_blocks=2, n_layers=1)


def test_get_glasflow_class_missing_prefix():
    with pytest.raises(ValueError, match=r"'glasflow' missing from name"):
        get_glasflow_class("realnvp")


def test_get_glasflow_class_invalid_flow():
    with pytest.raises(
        ValueError, match=r"invalid is not a known glasflow flow"
    ):
        get_glasflow_class("glasflow.invalid")


@pytest.mark.integration_test
def test_glasflow_integration(tmp_path):

    from glasflow.flows import RealNVP

    flow_config = dict(
        ftype="glasflow-realnvp",
        n_inputs=2,
    )

    flowmodel = FlowModel(flow_config=flow_config, output=tmp_path / "test")

    flowmodel.initialise()

    assert isinstance(flowmodel.model, GlasflowWrapper)
    assert isinstance(flowmodel.model._flow, RealNVP)

    flowmodel.train(np.random.randn(100, 2))



import matplotlib.pyplot as plt
import numpy as np
import os
import seaborn as sns

from ..flowmodel.clustering import ClusteringFlowModel
from ...livepoint import live_points_to_array
from ...model import Model
from ...plot import (
    nessai_style,
    plot_1d_comparison,
)
from ...proposal.flowproposal import FlowProposal


class ClusteringFlowProposal(FlowProposal):
    """FlowProposal that includes clustering for multi-modal live points.

    Parameters
    ----------
    model
        The model class used for computing the log-prior and log-likelihood.
    max_n_clusters
        The maximum number of clusters allowed. This is passed to
        :py:class:`~nessai.flowmodel.clustering.ClusteringFlowModel`. If not
        specified, the default in the flow model class will be used.
    kwargs
        Keyword arguments passed to
        :py:class:`~nessai.proposal.flowproposal.FlowProposal`, see that class
        for a complete list.
    """

    _FlowModelClass = ClusteringFlowModel

    def __init__(
        self, model: Model, max_n_clusters: int = None, **kwargs
    ) -> None:
        flow_config = kwargs.pop("flow_config", {}) or {}
        if max_n_clusters is not None:
            flow_config["max_n_clusters"] = max_n_clusters
        super().__init__(model, flow_config=flow_config, **kwargs)

    @nessai_style()
    def _plot_training_data(self, output: str) -> None:
        """Plot the training data and compare to the results.

        Plots each cluster individually.
        """
        z_training_data, _ = self.forward_pass(
            self.training_data, rescale=True
        )
        prime_array = live_points_to_array(
            self.training_data_prime,
            self.prime_parameters,
        )
        cluster_labels = self.flow.get_cluster_labels(prime_array)

        fig = plt.figure()
        plt.scatter(
            prime_array[:, 0],
            prime_array[:, 1],
            c=cluster_labels,
        )
        fig.savefig(os.path.join(output, "training_clusters"))
        plt.close(fig)

        z_gen = np.random.randn(self.training_data.size, self.dims)

        fig = plt.figure()
        plt.hist(np.sqrt(np.sum(z_training_data**2, axis=1)), "auto")
        plt.xlabel("Radius")
        fig.savefig(os.path.join(output, "radial_dist.png"))
        plt.close(fig)

        plot_1d_comparison(
            z_training_data,
            z_gen,
            labels=["z_live_points", "z_generated"],
            convert_to_live_points=True,
            filename=os.path.join(output, "z_comparison.png"),
        )

        x_prime_gen, log_prob = self.backward_pass(z_gen, rescale=False)
        gen_cluster_labels = self.flow.get_cluster_labels(
            live_points_to_array(x_prime_gen, self.prime_parameters)
        )
        x_prime_gen["logL"] = log_prob
        x_gen, log_J = self.inverse_rescale(x_prime_gen)
        (
            x_gen,
            log_J,
            x_prime_gen,
            gen_cluster_labels,
        ) = self.check_prior_bounds(
            x_gen, log_J, x_prime_gen, gen_cluster_labels
        )
        x_gen["logL"] += log_J

        labels = set(np.unique(gen_cluster_labels)) | set(
            np.unique(cluster_labels)
        )
        to_plot = [
            x_gen[np.squeeze(gen_cluster_labels == label)] for label in labels
        ]
        plot_labels = ["live_points"] + [f"cluster_{i}" for i in labels]
        colours = ["k"] + list(
            sns.color_palette("colorblind", n_colors=len(plot_labels) - 1),
        )

        plot_1d_comparison(
            self.training_data,
            *to_plot,
            parameters=self.model.names,
            labels=plot_labels,
            colours=colours,
            filename=os.path.join(output, "x_comparison.png"),
        )

        if self.parameters_to_rescale:
            plot_1d_comparison(
                self.training_data_prime,
                x_prime_gen,
                parameters=self.prime_parameters,
                labels=["live points", "generated"],
                filename=os.path.join(output, "x_prime_comparison.png"),
            )



from ..proposal.clustering import ClusteringFlowProposal
from ...gw.proposal import GWFlowProposal


class ClusteringGWFlowProposal(ClusteringFlowProposal, GWFlowProposal):
    """Augmented version of GWFlowProposal.

    See :obj:`~nessai.proposal.augmented.ClusteringFlowProposal` and
    :obj:`~nessai.gw.proposal.GWFlowPropsosal`
    """

    pass



from functools import partial
from glasflow.flows import CouplingNSF, RealNVP

from ...flows.base import BaseFlow

known_flows = {
    "nsf": CouplingNSF,
    "realnvp": RealNVP,
}


class GlasflowWrapper(BaseFlow):
    """Wrapper for glasflow flow classes"""

    def __init__(
        self,
        FlowClass,
        n_inputs,
        n_neurons,
        n_blocks,
        n_layers,
        **kwargs,
    ) -> None:
        super().__init__()

        n_conditional_inputs = kwargs.pop("context_features", None)
        self._flow = FlowClass(
            n_inputs=n_inputs,
            n_transforms=n_blocks,
            n_blocks_per_transform=n_layers,
            n_neurons=n_neurons,
            n_conditional_inputs=n_conditional_inputs,
            **kwargs,
        )

    def forward(self, x, context=None):
        return self._flow.forward(x, conditional=context)

    def inverse(self, z, context=None):
        return self._flow.inverse(z, conditional=context)

    def log_prob(self, x, context=None):
        return self._flow.log_prob(x, conditional=context)

    def sample(self, n, context=None):
        return self._flow.sample(n, conditional=context)

    def forward_and_log_prob(self, x, context=None):
        return self._flow.forward_and_log_prob(x, conditional=context)

    def sample_and_log_prob(self, n, context=None):
        return self._flow.sample_and_log_prob(n, conditional=context)

    def sample_latent_distribution(self, n, context=None):
        if context is not None:
            raise ValueError
        return self._flow.distribution.sample(n)

    def base_distribution_log_prob(self, z, context=None):
        if context is not None:
            raise ValueError("Context must be None")
        return self._flow.base_distribution_log_prob(z)

    def freeze_transform(self):
        self._flow._transform.requires_grad_(False)

    def unfreeze_transform(self):
        self._flow._transform.requires_grad_(True)


def get_glasflow_class(name):
    """Get the class for a glasflow flow.

    Note: the name must start with :code:`glasflow.`
    """
    name = name.lower()
    if "glasflow" not in name:
        raise ValueError("'glasflow' missing from name")
    short_name = name.replace("glasflow-", "")
    if short_name not in known_flows:
        raise ValueError(f"{name} is not a known glasflow flow")
    FlowClass = known_flows.get(short_name)
    return partial(GlasflowWrapper, FlowClass)



#!/usr/bin/env python

# Example of using the importance nested sampler with a non-uniform
# prior in the unit hypercube space.

import numpy as np
import os
from scipy.stats import norm, truncnorm

from nessai.flowsampler import FlowSampler
from nessai.model import Model
from nessai.utils import setup_logger
from nessai.plot import corner_plot

output = os.path.join("outdir", "ins_non_uniform_prior")
logger = setup_logger(output=output)


class ModelWithNonUniformPrior(Model):
    """A likelihood that has a non-uniform prior in the unit-hypercube"""

    def __init__(self, dims):
        self.names = [f"x_{d}" for d in range(dims)]
        self.bounds = {n: [-10.0, 10.0] for n in self.names}

        # Gaussian prior truncated on [-10, 10] with mean 0 and scale 0.5
        scale = 0.5
        self.prior_dist = truncnorm(-10 / scale, 10 / scale, scale=scale)

        # Define the distribution in the uniform hypercube
        # Will be centred at 0.5
        loc = 0.5
        h_scale = scale / 20
        self.hypercube_prior_dist = truncnorm(
            (0 - loc) / h_scale,  # Limits [0, 1]
            (1 - loc) / h_scale,
            loc=loc,
            scale=h_scale,
        )
        self.likelihood_dist = norm(loc=1.0, scale=0.5)

    def log_prior(self, x):
        """Log-prior"""
        log_p = np.log(self.in_bounds(x), dtype=float)
        log_p += self.prior_dist.logpdf(self.unstructured_view(x)).sum(axis=-1)
        return log_p

    def log_likelihood(self, x):
        "Log-likelihood"
        return self.likelihood_dist.logpdf(self.unstructured_view(x)).sum(
            axis=-1
        )

    def from_unit_hypercube(self, x):
        """Mapping from the unit hypercube.

        This mapping does not map from uniform to the correct prior.
        """
        x_out = x.copy()
        for n in self.names:
            x_out[n] = (self.bounds[n][1] - self.bounds[n][0]) * x[
                n
            ] + self.bounds[n][0]
        return x_out

    def log_prior_unit_hypercube(self, x) -> np.ndarray:
        """The log-prior defined in the unit hypercube space.

        This must be defined based on the how `from_unit_hypercube` is
        defined.
        """
        return np.log(
            self.in_unit_hypercube(x), dtype=float
        ) + self.hypercube_prior_dist.logpdf(self.unstructured_view(x)).sum(
            axis=-1
        )


# Run standard nessai for reference
model = ModelWithNonUniformPrior(2)
fs = FlowSampler(
    model,
    nlive=1000,
    output=os.path.join(output, "standard"),
    resume=False,
    seed=1234,
    importance_nested_sampler=False,
)
fs.run()

# Run the importance nested sampler
model = ModelWithNonUniformPrior(2)
fs_ins = FlowSampler(
    model,
    nlive=1000,
    output=os.path.join(output, "ins"),
    resume=False,
    seed=1234,
    importance_nested_sampler=True,
)
fs_ins.run()

# Compare the evidences
print(f"Log-evidences: {fs.log_evidence:.3f} vs {fs_ins.log_evidence:.3f}")

# Plot a comparison of the posteriors
fig = corner_plot(fs.posterior_samples, color="C0", include=model.names)
fig = corner_plot(
    fs_ins.posterior_samples,
    color="C1",
    fig=fig,
    include=model.names,
    filename=os.path.join(output, "comparison.png"),
)



from nessai.experimental.flowmodel.clustering import ClusteringFlowModel as CFM
import numpy as np
import pytest
from unittest.mock import create_autospec, patch


@pytest.fixture
def cfm():
    return create_autospec(CFM)


@pytest.mark.requires("faiss")
def test_init(cfm, tmp_path, caplog):
    caplog.set_level("DEBUG")
    flow_config = {}
    training_config = {}
    output = tmp_path
    with patch(
        "nessai.experimental.flowmodel.clustering.FlowModel.__init__"
    ) as mock_parent_init:
        CFM.__init__(
            cfm,
            flow_config=flow_config,
            training_config=training_config,
            output=output,
        )

    mock_parent_init.assert_called_once_with(
        flow_config=flow_config, training_config=training_config, output=output
    )
    assert "faiss version" in str(caplog.text)


@pytest.mark.requires("faiss")
@pytest.mark.integration_test
def test_clustering_integration(tmp_path, caplog):
    caplog.set_level("DEBUG")
    fm = CFM(flow_config=dict(n_inputs=2), output=tmp_path)
    n = 100
    samples = np.concatenate(
        [
            -10 * np.ones([n, 2]),
            10 * np.ones([n, 2]),
        ],
        axis=0,
    )
    fm.train_clustering(samples)
    assert fm.n_clusters == 2

    labels_pos = fm.get_cluster_labels(np.array([[10, 10]]))
    labels_neg = fm.get_cluster_labels(np.array([[-10, -10]]))
    assert labels_neg != labels_pos



import pytest

from nessai.flowsampler import FlowSampler
from nessai.experimental.proposal.clustering import ClusteringFlowProposal


@pytest.mark.requires("faiss")
@pytest.mark.slow_integration_test
def test_sampling_with_clusteringflowproposal(integration_model, tmp_path):
    fp = FlowSampler(
        integration_model,
        output=tmp_path / "clustering_sampling_test",
        resume=False,
        nlive=100,
        plot=False,
        maximum_uninformed=100,
        seed=1234,
        max_iteration=200,
        flow_class="clusteringflowproposal",
    )
    fp.run()
    assert isinstance(fp.ns.proposal, ClusteringFlowProposal)



# -*- coding: utf-8 -*-
"""
Utilities for configuring FlowModel.
"""
import copy
from warnings import warn

from . import config as default_config
from ..flows.utils import get_n_neurons


def update_flow_config(cfg):
    """Update the model (flow) configuration dictionary based on the defaults.

    Parameters
    ----------
    d : Union[dict, None]
        Dictionary with the current configuration. If None, then the default is
        used.

    Returns
    -------
    dict
        Updated configuration dictionary.

    Raises
    ------
    TypeError
        Raised if the input is not a dictionary or None.
    """
    default = default_config.flow.asdict()
    if cfg is None:
        return default
    elif not isinstance(cfg, dict):
        raise TypeError(
            "Must pass a dictionary to update the default model config"
        )
    default.update(copy.deepcopy(cfg))
    default["n_neurons"] = get_n_neurons(
        n_neurons=default.get("n_neurons"),
        n_inputs=default.get("n_inputs"),
    )
    return default


def update_model_config(cfg):
    msg = (
        "`update_model_config` is deprecated, use `update_flow_config` instead"
    )
    warn(msg, FutureWarning)
    return update_flow_config(cfg)


def update_training_config(cfg):
    default = default_config.training.asdict()
    if cfg is None:
        return default
    elif not isinstance(cfg, dict):
        raise TypeError(
            "Must pass a dictionary to update the default model config"
        )
    default.update(copy.deepcopy(cfg))
    if default["noise_type"] is not None and default["noise_scale"] is None:
        raise RuntimeError(
            "`noise_scale` must be specified when `noise_type` is given."
        )
    if isinstance(default["noise_scale"], float):
        if default["noise_type"] is None:
            default["noise_type"] = "constant"
    elif default["noise_scale"] is not None:
        raise TypeError(
            "`noise_scale` must be a float. "
            f"'Got type: {type(default['noise_scale'])}"
        )
    return default


def update_config(flow_config, training_config=None):
    """
    Update the configuration dictionary to include the defaults.

    Notes
    -----
    The default configuration is specified in :py:mod:`nessai.flowmodel.config`


    The kwargs can contain any additional keyword arguments that are specific
    to the type of flow being used.

    Parameters
    ----------
    flow_config : dict
        Dictionary with flow configuration
    training_config : dict
        Dictionary with training config

    Returns
    -------
    dict
        Dictionary with updated flow configuration
    dict
        Dictionary with updated training configuration
    """
    if flow_config is not None and (
        "model_config" in flow_config
        or set(flow_config.keys()).intersection(
            set(default_config.training.asdict().keys())
        )
    ):
        if "model_config" in flow_config:
            msg = (
                "Specifying `model_config` in `flow_config` is now deprecated."
                " Please specify `flow_config` and `training_config` instead."
            )
            warn(msg, FutureWarning)
        flow_config = copy.deepcopy(flow_config)
        flow_config.update(flow_config.pop("model_config", {}))

        if training_config is None:
            training_config = {}

        for key in default_config.training.asdict():
            if key in flow_config:
                warn(
                    (
                        f"Key `{key}` should now be specified in "
                        "`training_config`"
                    ),
                    FutureWarning,
                )
                if key in training_config:
                    raise RuntimeError(
                        f"`{key}` is already present in training config"
                    )
                training_config[key] = flow_config.pop(key)

    if flow_config is not None and "device_tag" in flow_config:
        msg = (
            "Specifying `device_tag` in `flow_config` is deprecated. "
            "It should now be specified in `training_config`."
        )
        warn(msg, FutureWarning)
        training_config = flow_config.pop("device_tag")

    if flow_config is not None and "inference_device_tag" in flow_config:
        msg = (
            "Specifying `inference_device_tag` in `flow_config` is deprecated."
            " It should now be specified in `training_config`."
        )
        warn(msg, FutureWarning)
        training_config = flow_config.pop("inference_device_tag")

    flow_config = update_flow_config(flow_config)
    training_config = update_training_config(training_config)
    return flow_config, training_config
